#############################################################################
# Copyright (c) 2018, Voilà Contributors                                    #
# Copyright (c) 2018, QuantStack                                            #
#                                                                           #
# Distributed under the terms of the BSD 3-Clause License.                  #
#                                                                           #
# The full license is in the file LICENSE, distributed with this software.  #
#############################################################################
import tornado.web

from ..handler import VoilaHandler


class TornadoVoilaHandler(VoilaHandler):
    @tornado.web.authenticated
    async def get(self, path=None):
        gen = self.get_generator(path=path)
        async for html in gen:
            self.write(html)
            self.flush()



#############################################################################
# Copyright (c) 2018, Voilà Contributors                                    #
# Copyright (c) 2018, QuantStack                                            #
#                                                                           #
# Distributed under the terms of the BSD 3-Clause License.                  #
#                                                                           #
# The full license is in the file LICENSE, distributed with this software.  #
#############################################################################

from ._version import __version__  # noqa
from .server_extension import _load_jupyter_server_extension  # noqa
from .server_extension import load_jupyter_server_extension  # noqa
import warnings

warnings.filterwarnings("default", category=DeprecationWarning, module="traitlets")


def _jupyter_nbextension_paths():
    return [
        {
            "section": "notebook",
            "src": "static",
            "dest": "voila",
            "require": "voila/extension",
        }
    ]


def _jupyter_labextension_paths():
    return [{"src": "labextension", "dest": "@voila-dashboards/jupyterlab-preview"}]



#############################################################################
# Copyright (c) 2018, Voilà Contributors                                    #
# Copyright (c) 2018, QuantStack                                            #
#                                                                           #
# Distributed under the terms of the BSD 3-Clause License.                  #
#                                                                           #
# The full license is in the file LICENSE, distributed with this software.  #
#############################################################################


import asyncio
import os
from pathlib import Path
from typing import Dict

from jupyter_server.base.handlers import JupyterHandler
from jupyter_server.utils import url_path_join
from nbclient.util import ensure_async
from tornado.httputil import split_host_and_port
from traitlets.traitlets import Bool

from .configuration import VoilaConfiguration

from ._version import __version__
from .notebook_renderer import NotebookRenderer
from .request_info_handler import RequestInfoSocketHandler
from .utils import ENV_VARIABLE, create_include_assets_functions, get_page_config


class BaseVoilaHandler(JupyterHandler):
    def initialize(self, **kwargs):
        self.voila_configuration = kwargs["voila_configuration"]

    def render_template(self, name, **ns):
        """Render the Voila HTML template, respecting the theme and nbconvert template."""
        template_arg = (
            self.get_argument("template", self.voila_configuration.template)
            if self.voila_configuration.allow_template_override == "YES"
            else self.voila_configuration.template
        )

        ns = {
            **ns,
            **self.template_namespace,
            **create_include_assets_functions(template_arg, self.base_url),
        }
        if "theme" not in ns:
            theme_arg = (
                self.get_argument("theme", self.voila_configuration.theme)
                if self.voila_configuration.allow_theme_override == "YES"
                else self.voila_configuration.theme
            )
            ns["theme"] = theme_arg

        template = self.get_template(name)
        return template.render(**ns)

    def get_template(self, name):
        """Return the jinja template object for a given name"""
        voila_env = self.settings["voila_jinja2_env"]
        template = voila_env.get_template(name)
        if template is not None:
            return template

        return self.settings["jinja2_env"].get_template(name)


class VoilaHandler(BaseVoilaHandler):
    def initialize(self, **kwargs):
        super().initialize(**kwargs)
        self.notebook_path = kwargs.pop("notebook_path", [])  # should it be []
        self.template_paths = kwargs.pop("template_paths", [])
        self.traitlet_config = kwargs.pop("config", None)
        self.voila_configuration: VoilaConfiguration = kwargs["voila_configuration"]
        self.prelaunch_hook = kwargs.get("prelaunch_hook", None)
        # we want to avoid starting multiple kernels due to template mistakes
        self.kernel_started = False

    async def get_generator(self, path=None):
        # if the handler got a notebook_path argument, always serve that
        notebook_path = self.notebook_path or path

        if (
            self.notebook_path and path
        ):  # when we are in single notebook mode but have a path
            self.redirect_to_file(path)
            return

        cwd = os.path.dirname(notebook_path)
        # Adding request uri to kernel env
        request_info = {}
        request_info[ENV_VARIABLE.SCRIPT_NAME] = self.request.path
        request_info[
            ENV_VARIABLE.PATH_INFO
        ] = ""  # would be /foo/bar if voila.ipynb/foo/bar was supported
        request_info[ENV_VARIABLE.QUERY_STRING] = str(self.request.query)
        request_info[ENV_VARIABLE.SERVER_SOFTWARE] = f"voila/{__version__}"
        request_info[ENV_VARIABLE.SERVER_PROTOCOL] = str(self.request.version)
        host, port = split_host_and_port(self.request.host.lower())

        request_info[ENV_VARIABLE.SERVER_PORT] = str(port) if port else ""
        request_info[ENV_VARIABLE.SERVER_NAME] = host

        # Add HTTP Headers as env vars following rfc3875#section-4.1.18
        if len(self.voila_configuration.http_header_envs) > 0:
            for header_name in self.request.headers:
                config_headers_lower = [
                    header.lower()
                    for header in self.voila_configuration.http_header_envs
                ]
                # Use case insensitive comparison of header names as per rfc2616#section-4.2
                if header_name.lower() in config_headers_lower:
                    env_name = f'HTTP_{header_name.upper().replace("-", "_")}'
                    request_info[env_name] = self.request.headers.get(header_name)

        template_arg = self.get_argument("template", None)
        theme_arg = self.get_argument("theme", None)

        # Compose reply
        self.set_header("Content-Type", "text/html")
        self.set_header("Cache-Control", "no-cache, no-store, must-revalidate")
        self.set_header("Pragma", "no-cache")
        self.set_header("Expires", "0")

        try:
            current_notebook_data: Dict = self.kernel_manager.notebook_data.get(
                notebook_path, {}
            )
            pool_size: int = self.kernel_manager.get_pool_size(notebook_path)
        except AttributeError:
            # For server extension case.
            current_notebook_data = {}
            pool_size = 0

        # Check if the conditions for using pre-heated kernel are satisfied.
        if self.should_use_rendered_notebook(
            current_notebook_data,
            pool_size,
            template_arg,
            theme_arg,
            self.request.arguments,
        ):
            # Get the pre-rendered content of notebook, the result can be all rendered cells
            # of the notebook or some rendred cells and a generator which can be used by this
            # handler to continue rendering calls.

            (
                render_task,
                rendered_cache,
                kernel_id,
            ) = await self.kernel_manager.get_rendered_notebook(
                notebook_name=notebook_path,
                extra_kernel_env_variables={
                    ENV_VARIABLE.VOILA_REQUEST_URL: self.request.full_url()
                },
            )

            RequestInfoSocketHandler.send_updates(
                {"kernel_id": kernel_id, "payload": request_info}
            )
            # Send rendered cell to frontend
            if len(rendered_cache) > 0:
                yield "".join(rendered_cache)

            # Wait for current running cell finish and send this cell to
            # frontend.
            rendered, rendering = await render_task
            if len(rendered) > len(rendered_cache):
                html_snippet = "".join(rendered[len(rendered_cache) :])
                yield html_snippet

            # Continue render cell from generator.
            async for html_snippet, _ in rendering:
                yield html_snippet

        else:
            # All kernels are used or pre-heated kernel is disabled, start a normal kernel.

            supported_file_extensions = [".ipynb"]
            supported_file_extensions.extend(
                [x.lower() for x in self.voila_configuration.extension_language_mapping]
            )
            file_extenstion = Path(notebook_path).suffix.lower()
            if file_extenstion not in supported_file_extensions:
                self.redirect_to_file(path)
                return
            mathjax_config = self.settings.get("mathjax_config")
            mathjax_url = self.settings.get("mathjax_url")
            gen = NotebookRenderer(
                request_handler=self,
                voila_configuration=self.voila_configuration,
                traitlet_config=self.traitlet_config,
                notebook_path=notebook_path,
                template_paths=self.template_paths,
                config_manager=self.config_manager,
                contents_manager=self.contents_manager,
                base_url=self.base_url,
                kernel_spec_manager=self.kernel_spec_manager,
                prelaunch_hook=self.prelaunch_hook,
                page_config=get_page_config(
                    base_url=self.base_url,
                    settings=self.settings,
                    log=self.log,
                    voila_configuration=self.voila_configuration,
                ),
                mathjax_config=mathjax_config,
                mathjax_url=mathjax_url,
            )

            await gen.initialize(template=template_arg, theme=theme_arg)

            def time_out():
                """If not done within the timeout, we send a heartbeat
                this is fundamentally to avoid browser/proxy read-timeouts, but
                can be used in a template to give feedback to a user
                """

                return "<script>window.voila_heartbeat()</script>\n"

            kernel_env = {**os.environ, **request_info}
            kernel_env[ENV_VARIABLE.VOILA_PREHEAT] = "False"
            kernel_env[ENV_VARIABLE.VOILA_BASE_URL] = self.base_url
            kernel_env[ENV_VARIABLE.VOILA_SERVER_URL] = self.settings.get(
                "server_url", "/"
            )
            kernel_env[ENV_VARIABLE.VOILA_REQUEST_URL] = self.request.full_url()
            kernel_env[ENV_VARIABLE.VOILA_APP_PORT] = request_info[
                ENV_VARIABLE.SERVER_PORT
            ]
            kernel_id = await ensure_async(
                self.kernel_manager.start_kernel(
                    kernel_name=gen.notebook.metadata.kernelspec.name,
                    path=cwd,
                    env=kernel_env,
                )
            )
            kernel_future = self.kernel_manager.get_kernel(kernel_id)
            queue = asyncio.Queue()

            async def put_html():
                async for html_snippet, _ in gen.generate_content_generator(
                    kernel_id, kernel_future
                ):
                    await queue.put(html_snippet)

                await queue.put(None)

            asyncio.ensure_future(put_html())

            # If not done within the timeout, we send a heartbeat
            # this is fundamentally to avoid browser/proxy read-timeouts, but
            # can be used in a template to give feedback to a user
            while True:
                try:
                    html_snippet = await asyncio.wait_for(
                        queue.get(), self.voila_configuration.http_keep_alive_timeout
                    )
                except asyncio.TimeoutError:
                    yield time_out()
                else:
                    if html_snippet is None:
                        break
                    yield html_snippet

    def redirect_to_file(self, path):
        self.redirect(url_path_join(self.base_url, "voila", "files", path))

    def should_use_rendered_notebook(
        self,
        notebook_data: Dict,
        pool_size: int,
        template_name: str,
        theme: str,
        request_args: Dict,
    ) -> Bool:
        if pool_size == 0:
            return False
        if len(notebook_data) == 0:
            return False

        rendered_template = notebook_data.get("template")
        rendered_theme = notebook_data.get("theme")
        if template_name is not None and template_name != rendered_template:
            return False
        if theme is not None and rendered_theme != theme:
            return False

        return True



#############################################################################
# Copyright (c) 2021, Voilà Contributors                                    #
# Copyright (c) 2021, QuantStack                                            #
#                                                                           #
# Distributed under the terms of the BSD 3-Clause License.                  #
#                                                                           #
# The full license is in the file LICENSE, distributed with this software.  #
#############################################################################


import os
import sys
import traceback
from functools import partial
from copy import deepcopy
from typing import Generator, List, Tuple, Union

import nbformat
import tornado.web
from jupyter_core.utils import ensure_async
from jupyter_server.config_manager import recursive_update
from nbclient.exceptions import CellExecutionError
from nbconvert.preprocessors.clearoutput import ClearOutputPreprocessor
from traitlets.config.configurable import LoggingConfigurable

from voila.configuration import VoilaConfiguration

from .execute import VoilaExecutor, strip_code_cell_warnings
from .exporter import VoilaExporter
from .paths import collect_template_paths
from .utils import ENV_VARIABLE


class NotebookRenderer(LoggingConfigurable):
    """Render the notebook into HTML string."""

    def __init__(self, **kwargs):
        super().__init__()
        self.request_handler = kwargs.get("request_handler")
        self.root_dir = kwargs.get("root_dir", [])
        self.notebook_path = kwargs.get("notebook_path", [])  # should it be []
        self.template_paths = kwargs.get("template_paths", [])
        self.traitlet_config = kwargs.get("traitlet_config", None)
        self.voila_configuration: VoilaConfiguration = kwargs.get("voila_configuration")
        self.config_manager = kwargs.get("config_manager")
        self.contents_manager = kwargs.get("contents_manager")
        self.kernel_spec_manager = kwargs.get("kernel_spec_manager")
        self.prelaunch_hook = kwargs.get("prelaunch_hook")
        self.base_url = kwargs.get("base_url")
        self.page_config = deepcopy(kwargs.get("page_config"))
        self.default_kernel_name = "python3"
        self.kernel_started = False
        self.stop_generator = False
        self.rendered_cache: List[str] = []
        self.mathjax_url = kwargs.get("mathjax_url")
        self.mathjax_config = kwargs.get("mathjax_config")

    async def initialize(self, **kwargs) -> None:
        """Initialize the notebook generator."""
        notebook_path = self.notebook_path
        self.notebook = await self.load_notebook(notebook_path)

        self.cwd = os.path.dirname(notebook_path)

        if self.prelaunch_hook:
            # Allow for preprocessing the notebook.
            # Can be used to add auth, do custom formatting/standardization
            # of the notebook, raise exceptions, etc
            #
            # Necessary inside of the handler if you need
            # to access the tornado request itself
            returned_notebook = self.prelaunch_hook(
                self.request_handler, notebook=self.notebook, cwd=self.cwd
            )
            if returned_notebook:
                self.notebook = returned_notebook

        _, basename = os.path.split(notebook_path)
        notebook_name = os.path.splitext(basename)[0]

        # we can override the template via notebook metadata or via
        # input parameter
        template_override = None
        if (
            "voila" in self.notebook.metadata
            and self.voila_configuration.allow_template_override in ["YES", "NOTEBOOK"]
        ):
            template_override = self.notebook.metadata["voila"].get("template")

        if self.voila_configuration.allow_template_override == "YES":
            template_arg = kwargs.get("template", None)
            template_override = (
                template_arg if template_arg is not None else template_override
            )
        if template_override:
            self.template_paths = collect_template_paths(
                ["voila", "nbconvert"], template_override
            )
        self.template_name = template_override or self.voila_configuration.template

        theme_override = self.voila_configuration.theme
        if (
            "voila" in self.notebook.metadata
            and self.voila_configuration.allow_theme_override in ["YES", "NOTEBOOK"]
        ):
            theme_override = self.notebook.metadata["voila"].get(
                "theme", theme_override
            )
        if self.voila_configuration.allow_theme_override == "YES":
            theme_arg = kwargs.get("theme", None)
            theme_override = theme_arg if theme_arg is not None else theme_override
        self.theme = theme_override
        # render notebook to html
        self.page_config["jupyterLabTheme"] = self.theme
        self.resources = {
            "base_url": self.base_url,
            "theme": self.theme,
            "template": self.template_name,
            "metadata": {"name": notebook_name},
            "page_config": self.page_config,
        }

        # include potential extra resources
        extra_resources = self.voila_configuration.config.VoilaConfiguration.resources
        # if no resources get configured from neither the CLI nor a config file,
        # extra_resources is a traitlets.config.loader.LazyConfigValue object
        # This seems to only happy with the notebook server and traitlets 5
        # Note that we use string checking for backward compatibility

        if "DeferredConfigString" in str(type(extra_resources)):
            from .configuration import VoilaConfiguration

            extra_resources = VoilaConfiguration.resources.from_string(extra_resources)
        if not isinstance(extra_resources, dict):
            extra_resources = extra_resources.to_dict()
        if extra_resources:
            recursive_update(self.resources, extra_resources)
        mathjax_full_url = f"{self.mathjax_url}?config={self.mathjax_config}"
        self.exporter = VoilaExporter(
            template_paths=self.template_paths,
            template_name=self.template_name,
            config=self.traitlet_config,
            contents_manager=self.contents_manager,  # for the image inlining
            theme=self.theme,  # we now have the theme in two places
            base_url=self.base_url,
            show_margins=self.voila_configuration.show_margins,
            mathjax_url=mathjax_full_url,
        )

        if self.voila_configuration.strip_sources:
            self.exporter.exclude_input = True
            self.exporter.exclude_output_prompt = True
            self.exporter.exclude_input_prompt = True

    def generate_content_generator(
        self,
        kernel_id: Union[str, None] = None,
        kernel_future=None,
    ) -> Generator:
        inner_kernel_start = partial(
            self._jinja_kernel_start, kernel_id=kernel_id, kernel_future=kernel_future
        )

        # kernel_start and cell_generator allow starting the kernel and executing the
        # notebook after (parts of) the template has been rendered and sent
        # to the client to allow progressive rendering.
        # Template should first call kernel_start, and then decide to use
        # notebook_executer cell_generator to implement progressive cell rendering.

        extra_context = {
            "frontend": "voila",
            "main_js": "voila.js",
            "kernel_start": inner_kernel_start,
            "cell_generator": self._jinja_cell_generator,
            "notebook_execute": self._jinja_notebook_execute,
        }
        # render notebook in snippets, then return an iterator so we can flush
        # them out to the browser progressively.
        return self.exporter.generate_from_notebook_node(
            self.notebook,
            resources=self.resources,
            extra_context=extra_context,
            page_config=self.page_config,
        )

    async def generate_content_hybrid(
        self,
        kernel_id: Union[str, None] = None,
        kernel_future=None,
    ) -> Tuple[List[str], Generator]:
        """Generate the HTML version of notebook, this process can be stopped
        anytime by setting `self.stop_generator=True`. The remaining cells can
        be rendered after by using the returned generator.
        """
        rendered = []
        generator = self.generate_content_generator(kernel_id, kernel_future)
        async for html_snippet, _ in generator:
            rendered.append(html_snippet)
            if self.stop_generator:
                break
            self.rendered_cache.append(html_snippet)
        return rendered, generator

    async def generate_content_str(
        self,
        kernel_id: Union[str, None] = None,
        kernel_future=None,
    ) -> str:
        """Generate the HTML version of notebook."""
        html = ""
        async for html_snippet, _ in self.generate_content_generator(
            kernel_id, kernel_future
        ):
            html += html_snippet
        return html

    async def _jinja_kernel_start(self, nb, kernel_id, kernel_future):
        assert not self.kernel_started, "kernel was already started"
        km = await ensure_async(kernel_future)
        self.executor = VoilaExecutor(
            nb,
            km=km,
            config=self.traitlet_config,
            show_tracebacks=self.voila_configuration.show_tracebacks,
        )

        self.executor.kc = await self.executor.async_start_new_kernel_client()

        # Set `VOILA_KERNEL_ID` environment variable, this variable help user can
        # identify which kernel the notebook use.
        if nb.metadata.kernelspec["language"] == "python":
            await ensure_async(
                self.executor.kc.execute(
                    f"""import os
                    \nos.environ["{ENV_VARIABLE.VOILA_KERNEL_ID}"]="{kernel_id}"
                    """,
                    store_history=False,
                )
            )

        self.kernel_started = True
        return kernel_id

    async def _jinja_notebook_execute(self, nb, kernel_id):
        result = await self.executor.async_execute(cleanup_kc=False)
        # we modify the notebook in place, since the nb variable cannot be
        # reassigned it seems in jinja2 e.g. if we do {% with nb = notebook_execute(nb, kernel_id) %}
        # ,the base template/blocks will not see the updated variable
        #  (it seems to be local to our block)
        nb.cells = result.cells

        await self._cleanup_resources()

    async def _jinja_cell_generator(self, nb, kernel_id):
        """Generator that will execute a single notebook cell at a time"""
        nb, _ = ClearOutputPreprocessor().preprocess(
            nb, {"metadata": {"path": self.cwd}}
        )
        for cell_idx, input_cell in enumerate(nb.cells):
            try:
                output_cell = await self.executor.execute_cell(
                    input_cell, None, cell_idx, store_history=False
                )
            except TimeoutError:
                output_cell = input_cell
                break
            except CellExecutionError:
                self.log.exception(
                    "Error at server while executing cell: %r", input_cell
                )
                if self.executor.should_strip_error():
                    strip_code_cell_warnings(input_cell)
                    self.executor.strip_code_cell_errors(input_cell)
                output_cell = input_cell
                break
            except Exception as e:
                self.log.exception(
                    "Error at server while executing cell: %r", input_cell
                )
                output_cell = nbformat.v4.new_code_cell()
                if self.executor.should_strip_error():
                    output_cell.outputs = [
                        {
                            "output_type": "stream",
                            "name": "stderr",
                            "text": "An exception occurred at the server (not the notebook). {}".format(
                                self.executor.cell_error_instruction
                            ),
                        }
                    ]
                else:
                    output_cell.outputs = [
                        {
                            "output_type": "error",
                            "ename": type(e).__name__,
                            "evalue": str(e),
                            "traceback": traceback.format_exception(*sys.exc_info()),
                        }
                    ]
            finally:
                yield output_cell

        await self._cleanup_resources()

    async def _cleanup_resources(self):
        await ensure_async(self.executor.kc.stop_channels())

    async def load_notebook(self, path):
        model = await ensure_async(self.contents_manager.get(path=path))
        if "content" not in model:
            raise tornado.web.HTTPError(404, f"{path} can not be found")
        __, extension = os.path.splitext(model.get("path", ""))
        if model.get("type") == "notebook":
            notebook = model["content"]
            notebook = await self.fix_notebook(notebook)
            return notebook
        elif extension in self.voila_configuration.extension_language_mapping:
            language = self.voila_configuration.extension_language_mapping[extension]
            notebook = await self.create_notebook(model, language=language)
            return notebook
        else:
            raise tornado.web.HTTPError(500, f"Failed to load {path}")

    async def fix_notebook(self, notebook):
        """Returns a notebook object with a valid kernelspec.

        In case the kernel is not found, we search for a matching kernel based on the language.
        """

        # Fetch kernel name from the notebook metadata
        if "kernelspec" not in notebook.metadata:
            notebook.metadata.kernelspec = nbformat.NotebookNode()
        kernelspec = notebook.metadata.kernelspec
        kernel_name = kernelspec.get("name", self.default_kernel_name)
        # We use `maybe_future` to support RemoteKernelSpecManager
        all_kernel_specs = await ensure_async(self.kernel_spec_manager.get_all_specs())
        # Find a spec matching the language if the kernel name does not exist in the kernelspecs
        if kernel_name not in all_kernel_specs:
            missing_kernel_name = kernel_name
            language = kernelspec.get(
                "language", notebook.metadata.get("language_info", {}).get("name", "")
            )
            kernel_name = await self.find_kernel_name_for_language(
                language.lower(), kernel_specs=all_kernel_specs
            )
            self.log.warning(
                "Could not find a kernel named %r, will use  %r",
                missing_kernel_name,
                kernel_name,
            )
        # We make sure the notebook's kernelspec is correct
        notebook.metadata.kernelspec.name = kernel_name
        notebook.metadata.kernelspec.display_name = all_kernel_specs[kernel_name][
            "spec"
        ]["display_name"]
        notebook.metadata.kernelspec.language = all_kernel_specs[kernel_name]["spec"][
            "language"
        ]
        return notebook

    async def create_notebook(self, model, language):
        all_kernel_specs = await ensure_async(self.kernel_spec_manager.get_all_specs())
        kernel_name = await self.find_kernel_name_for_language(
            language, kernel_specs=all_kernel_specs
        )
        spec = all_kernel_specs[kernel_name]
        notebook = nbformat.v4.new_notebook(
            metadata={
                "kernelspec": {
                    "display_name": spec["spec"]["display_name"],
                    "language": spec["spec"]["language"],
                    "name": kernel_name,
                }
            },
            cells=[nbformat.v4.new_code_cell(model["content"])],
        )
        return notebook

    async def find_kernel_name_for_language(self, kernel_language, kernel_specs=None):
        """Finds a best matching kernel name given a kernel language.

        If multiple kernels matches are found, we try to return the same kernel name each time.
        """
        if kernel_language in self.voila_configuration.language_kernel_mapping:
            return self.voila_configuration.language_kernel_mapping[kernel_language]
        if kernel_specs is None:
            kernel_specs = await ensure_async(self.kernel_spec_manager.get_all_specs())
        matches = [
            name
            for name, kernel in kernel_specs.items()
            if kernel["spec"]["language"].lower() == kernel_language.lower()
        ]
        if matches:
            # Sort by display name to get the same kernel each time.
            matches.sort(key=lambda name: kernel_specs[name]["spec"]["display_name"])
            return matches[0]
        else:
            raise tornado.web.HTTPError(
                500, "No Jupyter kernel for language %r found" % kernel_language
            )



#############################################################################
# Copyright (c) 2018, Voilà Contributors                                    #
# Copyright (c) 2018, QuantStack                                            #
#                                                                           #
# Distributed under the terms of the BSD 3-Clause License.                  #
#                                                                           #
# The full license is in the file LICENSE, distributed with this software.  #
#############################################################################

import mimetypes
from typing import Optional

import traitlets
from traitlets.config import Config

try:
    from jinja2 import pass_context
except ImportError:
    from jinja2 import contextfilter as pass_context

from nbconvert.exporters.html import HTMLExporter
from nbconvert.exporters.templateexporter import TemplateExporter
from nbconvert.filters.highlight import Highlight2HTML
from nbconvert.filters.markdown_mistune import (
    IPythonRenderer,
    MarkdownWithMath,
)
from .static_file_handler import TemplateStaticFileHandler
from .utils import create_include_assets_functions

try:
    from nbconvert.filters.markdown_mistune import MISTUNE_V3  # noqa

    NB_CONVERT_760 = True
except ImportError:
    NB_CONVERT_760 = False


class VoilaMarkdownRenderer(IPythonRenderer):
    """Custom markdown renderer that inlines images"""

    def __init__(self, contents_manager, *args, **kwargs):
        self.contents_manager = contents_manager
        super().__init__(*args, **kwargs)

    def image(self, text: str, url: str, title: Optional[str] = None):
        contents_manager = self.contents_manager
        # for nbconvert <7.6.0, the first argument is the URL
        src = url if NB_CONVERT_760 else text

        if contents_manager.file_exists(src):
            content = contents_manager.get(src, format="base64")
            data = content["content"].replace("\n", "")  # remove the newline
            mime_type, encoding = mimetypes.guess_type(src)
            src = f"data:{mime_type};base64,{data}"
        if NB_CONVERT_760:
            return super().image(text, src, title)
        else:
            return super().image(src, url, title)


class VoilaExporter(HTMLExporter):
    """Custom HTMLExporter that inlines the images using VoilaMarkdownRenderer"""

    base_url = traitlets.Unicode(help="Base url for resources").tag(config=True)

    markdown_renderer_class = traitlets.Type(VoilaMarkdownRenderer).tag(config=True)

    # Can be a ContentsManager from notebook or jupyter_server, so Any will have to do for now
    contents_manager = traitlets.Any()

    show_margins = traitlets.Bool(
        True,
        help='show left and right margins for the "lab" template, this gives a "classic" template look',
    ).tag(config=True)

    @traitlets.validate("template_name")
    def _validate_template_name(self, template_name):
        if template_name.value == "classic":
            self.log.warning(
                '"classic" template support will be removed in Voila 1.0.0, '
                'please use the "lab" template instead with the "--show-margins" '
                "option for a similar look"
            )
        return template_name.value

    @pass_context
    def markdown2html(self, context, source):
        # The voila exporter overrides the markdown renderer from the HTMLExporter
        # to inline images.
        cell = context["cell"]
        attachments = cell.get("attachments", {})
        cls = self.markdown_renderer_class
        renderer = cls(
            escape=False,
            attachments=attachments,
            contents_manager=self.contents_manager,
            anchor_link_text=self.anchor_link_text,
        )
        return MarkdownWithMath(renderer=renderer).render(source)

    # The voila exporter disables the CSSHTMLHeaderPreprocessor from the HTMLExporter.

    @property
    def default_config(self):
        c = Config(
            {
                "VoilaExporter": {
                    "markdown_renderer_class": "voila.exporter.VoilaMarkdownRenderer"
                }
            }
        )
        c.merge(super().default_config)
        return c

    # Overriding the default template file.

    @traitlets.default("template_file")
    def default_template_file(self):
        return "index.html.j2"

    async def generate_from_notebook_node(
        self, nb, resources=None, extra_context={}, page_config={}, **kw
    ):
        # this replaces from_notebook_node, but calls template.generate instead of template.render
        langinfo = nb.metadata.get("language_info", {})
        lexer = langinfo.get("pygments_lexer", langinfo.get("name", None))
        highlight_code = self.filters.get(
            "highlight_code", Highlight2HTML(pygments_lexer=lexer, parent=self)
        )
        self.register_filter("highlight_code", highlight_code)

        # We need outputs to be marked as trusted for the JupyterLab mimerendering logic
        def trusted_output(output):
            output["trusted"] = True
            return output

        self.register_filter("trusted", trusted_output)

        # NOTE: we don't call HTML or TemplateExporter' from_notebook_node
        nb_copy, resources = super(TemplateExporter, self).from_notebook_node(
            nb, resources, **kw
        )
        resources.setdefault("raw_mimetypes", self.raw_mimetypes)
        resources["global_content_filter"] = {
            "include_code": not self.exclude_code_cell,
            "include_markdown": not self.exclude_markdown,
            "include_raw": not self.exclude_raw,
            "include_unknown": not self.exclude_unknown,
            "include_input": not self.exclude_input,
            "include_output": not self.exclude_output,
            "include_input_prompt": not self.exclude_input_prompt,
            "include_output_prompt": not self.exclude_output_prompt,
            "no_prompt": self.exclude_input_prompt and self.exclude_output_prompt,
        }
        async for output in self.template.generate_async(
            nb=nb_copy,
            resources=resources,
            **extra_context,
            static_url=self.static_url,
            page_config=page_config,
        ):
            yield (output, resources)

    @property
    def environment(self):
        # enable Jinja async template execution
        self.enable_async = True
        env = super().environment
        if "jinja2.ext.do" not in env.extensions:
            env.add_extension("jinja2.ext.do")
        return env

    def get_template_paths(self):
        return self.template_path

    def static_url(self, path):
        """Mimics tornado.web.RequestHandler.static_url"""
        settings = {
            "static_url_prefix": f"{self.base_url}voila/templates/",
            "static_path": None,  # not used in TemplateStaticFileHandler.get_absolute_path
        }
        return TemplateStaticFileHandler.make_static_url(
            settings, f"{self.template_name}/static/{path}"
        )

    def _init_resources(self, resources):
        resources = super()._init_resources(resources)

        include_assets_functions = create_include_assets_functions(
            self.template_name, self.base_url
        )

        resources.update(include_assets_functions)

        resources["show_margins"] = self.show_margins

        return resources



#############################################################################
# Copyright (c) 2018, Voilà Contributors                                    #
# Copyright (c) 2018, QuantStack                                            #
#                                                                           #
# Distributed under the terms of the BSD 3-Clause License.                  #
#                                                                           #
# The full license is in the file LICENSE, distributed with this software.  #
#############################################################################

from nbclient.client import NotebookClient
from nbclient.exceptions import CellExecutionError
from nbconvert.preprocessors.clearoutput import ClearOutputPreprocessor
from traitlets import Bool, Unicode, Integer


def strip_code_cell_warnings(cell):
    """Strip any warning outputs and traceback from a code cell."""
    if cell["cell_type"] != "code":
        return cell

    outputs = cell["outputs"]

    cell["outputs"] = [
        output
        for output in outputs
        if output["output_type"] != "stream" or output["name"] != "stderr"
    ]

    return cell


class VoilaExecutor(NotebookClient):
    """Execute, but respect the output widget behaviour"""

    cell_error_instruction = Unicode(
        "Please run Voilà with --show_tracebacks=True or --debug to see the error message, or configure VoilaConfiguration.show_tracebacks.",
        config=True,
        help=("instruction given to user to debug cell errors"),
    )

    cell_timeout_instruction = Unicode(
        "Please run Voilà with --VoilaExecutor.interrupt_on_timeout=True to continue executing the rest of the notebook.",
        config=True,
        help=("instruction given to user to continue execution on timeout"),
    )

    show_tracebacks = Bool(
        False,
        config=True,
        help=("Whether to send tracebacks to clients on exceptions."),
    )

    startup_timeout: int = Integer(
        60,
        config=True,
        help=(
            """
            The time to wait (in seconds) for the kernel to start.
            If kernel startup takes longer, a RuntimeError is
            raised.
            """
        ),
    )

    def execute(self, nb, resources, km=None):
        try:
            result = super().execute()
        except CellExecutionError as e:
            self.log.error(e)
            result = (nb, resources)

        # Strip errors and traceback if not in debug mode
        if self.should_strip_error():
            self.strip_notebook_errors(nb)

        return result

    async def execute_cell(self, cell, resources, cell_index, store_history=True):
        try:
            result = await self.async_execute_cell(cell, cell_index, store_history)
        except TimeoutError as e:
            self.log.error(e)
            self.show_code_cell_timeout(cell)
            raise e

        # Strip errors and traceback if not in debug mode
        if self.should_strip_error():
            strip_code_cell_warnings(cell)
            self.strip_code_cell_errors(cell)

        return result

    def should_strip_error(self):
        """Return True if errors should be stripped from the Notebook, False otherwise, depending on the current config."""
        return not self.show_tracebacks

    def strip_notebook_errors(self, nb):
        """Strip error messages and traceback from a Notebook."""
        cells = nb["cells"]

        code_cells = [cell for cell in cells if cell["cell_type"] == "code"]

        for cell in code_cells:
            strip_code_cell_warnings(cell)
            self.strip_code_cell_errors(cell)

        return nb

    def strip_code_cell_errors(self, cell):
        """Strip any error outputs and traceback from a code cell."""
        # There is no 'outputs' key for markdown cells
        if cell["cell_type"] != "code":
            return cell

        outputs = cell["outputs"]

        error_outputs = [
            output for output in outputs if output["output_type"] == "error"
        ]

        error_message = "There was an error when executing cell [{}]. {}".format(
            cell["execution_count"], self.cell_error_instruction
        )

        for output in error_outputs:
            output["ename"] = "ExecutionError"
            output["evalue"] = "Execution error"
            output["traceback"] = [error_message]

        return cell

    def show_code_cell_timeout(self, cell):
        """Show a timeout error output in a code cell."""

        timeout_message = f"Cell execution timed out, aborting notebook execution. {self.cell_timeout_instruction}"

        output = {
            "output_type": "error",
            "ename": "TimeoutError",
            "evalue": "Timeout error",
            "traceback": [timeout_message],
        }

        cell["outputs"] = [output]


def executenb(nb, cwd=None, km=None, **kwargs):
    resources = {}
    if cwd is not None:
        resources["metadata"] = {"path": cwd}  # pragma: no cover
    # Clear any stale output, in case of exception
    nb, resources = ClearOutputPreprocessor().preprocess(nb, resources)
    executor = VoilaExecutor(nb, km=km, **kwargs)
    return executor.execute(nb, resources, km=km)
